/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "surface.pmh"

#define FE_SOBJ_CACHE_DEBUG		FALSE
#define FE_SOBJ_FILTER			TRUE

namespace fe
{

SurfaceOBJ::SurfaceOBJ(void)
{
	feLog("SurfaceOBJ()\n");
}

void SurfaceOBJ::initialize(void)
{
	feLog("SurfaceOBJ:initialize\n");
}

Protectable* SurfaceOBJ::clone(Protectable* pInstance)
{
	feLog("SurfaceOBJ::clone\n");

	SurfaceOBJ* pSurfaceOBJ= pInstance?
			dynamic_cast<SurfaceOBJ*>(pInstance): new SurfaceOBJ();

	checkCache();

	SurfaceSphere::clone(pSurfaceOBJ);

	//* copy attributes

	return pSurfaceOBJ;
}

/***
# OBJ file created by ply_to_obj.c
#
g Object001

v  40.6266  28.3457  -1.10804
v  40.0714  30.4443  -1.10804

vn  -0.966742  -0.255752  9.97231e-09
vn  -0.966824  0.255443  3.11149e-08

f  7  6  1
f  1  2  7
*/

void SurfaceOBJ::cache(void)
{
	feLog("SurfaceOBJ::cache\n");

	if(!m_record.isValid())
	{
		feLog("SurfaceOBJ::cache record invalid\n");
		return;
	}

	SurfaceTriangles::cache();

	SurfaceFile surfaceFileRV;
	surfaceFileRV.bind(m_record);

	String path=registry()->master()->catalog()->catalog<String>("path:media")+
			"/"+surfaceFileRV.filename();

	std::ifstream stream;

	stream.open(path.c_str(),std::fstream::in);
	if(!stream.good())
	{
		feLog("SurfaceOBJ::cache failed to open \"%s\"\n",path.c_str());
		return;
	}

	U32 file_vertices=0;
	U32 file_normals=0;
	U32 file_faces=0;
	char line[256];
	char type[256];
	char face0[256];
	char face1[256];
	char face2[256];

	//* first pass: just count
	while(stream.good())
	{
		type[0]=0;
		stream.getline(line,256);
		sscanf(line,"%s",type);
		if(!strcmp(type,"g"))
		{
			char objname[256];
			sscanf(line,"%8s%64s",type,objname);
			setName(objname);
		}
		if(!strcmp(type,"v"))
		{
			file_vertices++;
		}
		if(!strcmp(type,"vn"))
		{
			file_normals++;
		}
		if(!strcmp(type,"f"))
		{
			file_faces++;
		}
	}

	//* TODO use a mesh, not discrete triangles
	m_vertices=file_faces*3;

	feLog("SurfaceOBJ::cache file_vertices=%d file_normals=%d file_faces=%d\n",
			file_vertices,file_normals,file_faces);

	if(file_vertices<file_normals)
	{
		feLog("SurfaceOBJ::cache missing vertices will be zeroed\n");
	}
	if(file_normals<file_vertices)
	{
		feLog("SurfaceOBJ::cache missing normals will be set to unit Z\n");
	}

	SpatialVector* pVertices=new SpatialVector[file_vertices];
	SpatialVector* pNormals=new SpatialVector[file_normals];

	m_pVertexArray=(SpatialVector*)reallocate(m_pVertexArray,
			m_vertices*sizeof(SpatialVector));
	m_pNormalArray=(SpatialVector*)reallocate(m_pNormalArray,
			m_vertices*sizeof(SpatialVector));

	//* back to beginning
	stream.clear();
	stream.seekg(0,std::ios::beg);

	U32 vertex=0;
	U32 file_vertex=0;
	U32 file_normal=0;
	U32 file_face=0;

	//* second pass: convert
	while(!stream.eof())
	{
		stream.getline(line,256);
		sscanf(line,"%8s",type);
		if(!strcmp(type,"v") && file_vertex<file_vertices)
		{
			float x;
			float y;
			float z;
			sscanf(line,"%8s%f%f%f",type,&x,&y,&z);

			SpatialVector& rFileVertex=pVertices[file_vertex++];
//			set(rFileVertex,z,x,y);
			set(rFileVertex,x,-z,y);

#if FE_SOBJ_CACHE_DEBUG
			feLog("vertex %d %s\n",file_vertex-1,c_print(rFileVertex));
#endif
		}
		if(!strcmp(type,"vn") && file_normal<file_normals)
		{
			float x;
			float y;
			float z;
			sscanf(line,"%8s%f%f%f",type,&x,&y,&z);

			SpatialVector& rFileNormal=pNormals[file_normal++];
//			set(rFileNormal,z,x,y);
			set(rFileNormal,x,-z,y);

#if FE_SOBJ_CACHE_DEBUG
			feLog("normal %d %s\n",file_normal-1,c_print(rFileNormal));
#endif
		}
		if(!strcmp(type,"f") && file_face<file_faces && vertex+2<m_vertices)
		{
			Vector3i vid;
			Vector3i tid;
			Vector3i nid;
			sscanf(line,"%8s%s%s%s",type,face0,face1,face2);

			set(vid);
			set(tid);
			set(nid);
			sscanf(face0,"%d/%d/%d",&vid[0],&tid[0],&nid[0]);
			sscanf(face1,"%d/%d/%d",&vid[1],&tid[1],&nid[1]);
			sscanf(face2,"%d/%d/%d",&vid[2],&tid[2],&nid[2]);

#if FE_SOBJ_CACHE_DEBUG
			feLog("face at %d: %d %d %d\n",vertex,vid[0],vid[1],vid[2]);
#endif

			if(!vid[0] || !vid[1] || !vid[2] ||
					vid[0]==vid[1] || vid[0]==vid[2] || vid[1]==vid[2])
			{
				feLog("SurfaceOBJ::cache bad face at %d: %d %d %d\n",
						vertex,vid[0],vid[1],vid[2]);
				continue;
			}

			BWORD bad=FALSE;
			for(U32 m=0;m<3 && !bad;m++)
			{
				const U32 vv=vid[m]-1;
				m_pVertexArray[vertex]=(vv<file_vertices)?
						pVertices[vv]: SpatialVector(0.0,0.0,0.0);
				const U32 nn=(nid[m])? nid[m]-1: vv;
				m_pNormalArray[vertex]=(nn<file_normals)?
						pNormals[nn]: SpatialVector(0.0,0.0,1.0);

#if FE_SOBJ_CACHE_DEBUG
				feLog(" v %s\n",c_print(m_pVertexArray[vertex]));
				feLog(" n %s\n",c_print(m_pNormalArray[vertex]));
#endif

#if FE_SOBJ_FILTER
				if(m)
				{
					const U32 vv2=vid[m-1]-1;
					SpatialVector other=(vv2<file_vertices)?
							pVertices[vv2]: SpatialVector(0.0,0.0,0.0);
					const float dist2 =
						magnitudeSquared(other-m_pVertexArray[vertex]);
					if(dist2<1e-6)
					{
						feLog("SurfaceOBJ::cache face at %d"
								" coincident verts %d %d  %s  %s\n",
								vertex,vv+1,vv2+1,
								c_print(other),
								c_print(m_pVertexArray[vertex]));
						bad=TRUE;
					}
				}
#endif
				if(!bad)
				{
					vertex++;
				}
			}
		}
	}

	stream.close();

	if(m_vertices!=vertex)
	{
		m_vertices=vertex;
		m_pVertexArray=(SpatialVector*)reallocate(m_pVertexArray,
				m_vertices*sizeof(SpatialVector));
		m_pNormalArray=(SpatialVector*)reallocate(m_pNormalArray,
				m_vertices*sizeof(SpatialVector));
	}

	feLog("SurfaceOBJ::cache name %s verts %d\n",verboseName().c_str(),vertex);
}

} // namespace
